import json
import os
import time
import sys
import requests
import logging

# Set up logging
logging.basicConfig(
    level=logging.INFO,
    format="%(asctime)s - %(levelname)s - %(funcName)s - %(message)s",
)


HEADERS = {"accept": "application/json", "Content-Type": "application/json"}

# Please make sure the `API_URLS` variable in the script matches the URLs provided to
# you at signup.
API_HOST = "1qbit.com"
API_URLS = {
    "upload_url": f"https://qarrot.{API_HOST}/api/upload",
    "decomposer": f"https://skd.{API_HOST}/",
    "optimizer": f"https://transpiler.{API_HOST}/",
    "scheduler": f"https://compiler.{API_HOST}/",
    "ftqc_emulator": f"https://ftqc.{API_HOST}/",
    "pre": f"https://pre.{API_HOST}/",
    "qare": f"https://qarrot.{API_HOST}/api/qare/",
}

RETRY_LIMIT = 5
# Please remember to update the `SAMPLE_PAYLOAD_FOLDER` variable to your own sample
# payload folder path.
SAMPLE_PAYLOAD_FOLDER = "/sample_payload"
PAYLOAD_PATHS = {
    "decomposer": os.path.join(SAMPLE_PAYLOAD_FOLDER, "decomposer_payload.json"),
    "optimizer": os.path.join(SAMPLE_PAYLOAD_FOLDER, "optimizer_payload.json"),
    "scheduler": os.path.join(SAMPLE_PAYLOAD_FOLDER, "scheduler_payload.json"),
    "ftqc_emulator": os.path.join(SAMPLE_PAYLOAD_FOLDER, "ftqc_emulator_payload.json"),
    "pre": os.path.join(SAMPLE_PAYLOAD_FOLDER, "pre_payload.json"),
    "qare": os.path.join(SAMPLE_PAYLOAD_FOLDER, "qare_payload.json"),
}
OUTPUT_DIRECTORY = "output"


def get_request_helper(url):
    """Helper function to send GET requests with retry logic."""
    retries = 0
    while retries <= RETRY_LIMIT:
        response = requests.get(url, headers=HEADERS)
        result = response.json()

        status = result.get("status")
        if status == "done":
            return result
        if status == "failed":
            logging.error("Request failed")
            return result

        wait = (retries + 1) * 10
        logging.info(
            "Request still processing. Waiting %s secs before retrying...", wait
        )
        time.sleep(wait)
        retries += 1

    logging.error("Request failed after %s retries", RETRY_LIMIT)
    return result


def run_decomposer(circuit_path):
    """Run the decomposer."""
    logging.info("Decomposer circuit path: %s", circuit_path)

    with open(PAYLOAD_PATHS["decomposer"], "r") as file:
        decomposer_request_body = json.load(file)
    decomposer_request_body["circuit_path"] = circuit_path
    decomposer_api_url = f"{API_URLS['decomposer']}decompose"

    logging.debug("Decomposer request body: %s", decomposer_request_body)

    # Send POST request
    post_decomposer_response = requests.post(
        url=decomposer_api_url, json=decomposer_request_body
    )
    request_id = post_decomposer_response.json().get("request_id")
    decomposer_get_request_url = f"{decomposer_api_url}/{request_id}"

    # Send GET request
    response = get_request_helper(decomposer_get_request_url)
    logging.debug("Decomposer response: %s", response)

    return response["sk_circuit_path"], response["accumulated_error"]


def run_optimizer(post_decomposer_circuit_path, sk_accumulated_error):
    """Run the optimizer."""
    logging.info("Post decomposer circuit path: %s", post_decomposer_circuit_path)
    with open(PAYLOAD_PATHS["optimizer"], "r") as file:
        optimizer_request_body = json.load(file)
    optimizer_request_body["file_path"] = post_decomposer_circuit_path
    optimizer_api_url = f"{API_URLS['optimizer']}transpile"

    logging.debug("Optimizer request body: %s", optimizer_request_body)

    # Send POST request
    response = requests.post(url=optimizer_api_url, json=optimizer_request_body)
    request_id = response.json().get("request_id")
    optimizer_get_request_url = f"{optimizer_api_url}/{request_id}"

    # Send GET request
    get_optimizer_response = get_request_helper(optimizer_get_request_url)
    logging.debug("Optimizer response: %s", get_optimizer_response)

    # Get Post optimizer circuit_path
    transpiled_circuit_path = get_optimizer_response["transpiled_circuit_path"]

    # Format circuit details
    circuit_details = get_optimizer_response
    for key in [
        "status",
        "request_id",
        "transpiled_circuit_path",
        "elapsed_time",
        "bypass_optimization",
        "message",
    ]:
        circuit_details.pop(key, None)

    circuit_details["sk_accumulated_error"] = sk_accumulated_error
    return transpiled_circuit_path, circuit_details


def run_scheduler(transpiled_circuit_path):
    """Run the scheduler."""
    logging.info("Transpiled circuit path: %s", transpiled_circuit_path)

    with open(PAYLOAD_PATHS["scheduler"], "r") as file:
        scheduler_request_body = json.load(file)
    scheduler_request_body["circuit_path"] = transpiled_circuit_path
    scheduler_api_url = f"{API_URLS['scheduler']}compile"

    logging.debug("Scheduler request body: %s", scheduler_request_body)

    # Send POST request
    response = requests.post(url=scheduler_api_url, json=scheduler_request_body)
    request_id = response.json().get("request_id")
    scheduler_get_request_url = f"{scheduler_api_url}/{request_id}"

    # Send GET request
    get_scheduler_response = get_request_helper(scheduler_get_request_url)
    logging.debug("Scheduler response: %s", get_scheduler_response)

    # Format layout details
    layout_metadata = get_scheduler_response["layout_metadata"]
    layout_details = {
        key: layout_metadata[key]
        for key in ["type", "layout_filename", "num_buffers", "num_logical_qubits"]
    }

    # Format schedule details
    schedule_details = {
        "schedule_metadata": get_scheduler_response["schedule_metadata"],
        "schedule_path": get_scheduler_response["schedule_filepath"],
    }
    return layout_details, layout_metadata, schedule_details


def run_ftqc_emulator():
    """Run the FTQC emulator."""
    with open(PAYLOAD_PATHS["ftqc_emulator"], "r") as file:
        ftqc_emulator_request_body = json.load(file)
    ftqc_emulator_api_url = f"{API_URLS['ftqc_emulator']}emulate"

    logging.debug("FTQC emulator request body: %s", ftqc_emulator_request_body)

    # Send POST request
    response = requests.post(
        ftqc_emulator_api_url, headers=HEADERS, json=ftqc_emulator_request_body
    )
    request_id = response.json().get("request_id")
    ftqc_emulator_get_request_url = f"{ftqc_emulator_api_url}/{request_id}"

    # Send GET request
    get_ftqc_emulator_response = get_request_helper(ftqc_emulator_get_request_url)
    logging.debug("FTQC emulator response: %s", get_ftqc_emulator_response)

    emulator_plot_path = get_ftqc_emulator_response["emulator_plot_path"]
    # Format ftqc_emulator details
    ftqc_emulator_emulation_details = get_ftqc_emulator_response
    for key in [
        "request_id",
        "status",
        "message",
        "elapsed_time",
        "emulator_plot_path",
    ]:
        ftqc_emulator_emulation_details.pop(key, None)

    ftqc_emulator_input_params = ftqc_emulator_request_body["qubit_technology"][
        "parameters"
    ]

    ftqc_emulator_inputs_for_pre = {
        "code": ftqc_emulator_request_body["code"],
        "qubit_technology": {
            "physical_fidelity": ftqc_emulator_input_params["hadamard_gate_fidelity"][
                "value"
            ],
            "preparation_fidelity": ftqc_emulator_input_params["preparation_fidelity"][
                "value"
            ],
            "t_gate_fidelity": ftqc_emulator_input_params["t_gate_fidelity"]["value"],
            "qubit_measurement_time": ftqc_emulator_input_params[
                "qubit_measurement_time"
            ],
        },
    }

    ftqc_emulator_inputs_for_pre.update(ftqc_emulator_emulation_details)
    return (
        ftqc_emulator_request_body,
        emulator_plot_path,
        ftqc_emulator_emulation_details,
        ftqc_emulator_inputs_for_pre,
    )


def run_pre(
    circuit_details, layout_details, ftqc_emulator_inputs_for_pre, schedule_details
):
    """Run the Physical Resource Estimator."""
    logging.info("Circuit details: %s", circuit_details)
    with open(PAYLOAD_PATHS["pre"], "r") as file:
        pre_request_body = json.load(file)

    pre_request_body.update(
        {
            "circuit": circuit_details,
            "layout": layout_details,
            "ftqc_emulator_emulation": ftqc_emulator_inputs_for_pre,
            "schedule": schedule_details,
        }
    )

    pre_api_url = f"{API_URLS['pre']}estimate"
    logging.debug("PRE request body: %s", pre_request_body)

    # Send POST request
    response = requests.post(pre_api_url, headers=HEADERS, json=pre_request_body)
    request_id = response.json().get("request_id")
    pre_get_request_url = f"{pre_api_url}/{request_id}"

    # Send GET request
    get_pre_response = get_request_helper(pre_get_request_url)
    logging.debug("PRE response: %s", get_pre_response)

    # Format PRE details
    pre_details = get_pre_response
    for key in ["status", "request_id", "elapsed_time", "message"]:
        pre_details.pop(key, None)
    return pre_details


def run_qare(
    circuit_details,
    layout_metadata,
    ftqc_emulator_request_body,
    emulator_plot_path,
    ftqc_emulator_emulation_details,
    pre_details,
):
    """Run the Quantum Architecure & Resource Estimation Report Generator."""
    with open(PAYLOAD_PATHS["qare"], "r") as file:
        qare_request_body = json.load(file)

    ftqc_emulator_details = ftqc_emulator_request_body

    ftqc_emulator_details.pop("number_of_cores", None)
    ftqc_emulator_details.update(ftqc_emulator_emulation_details)
    ftqc_emulator_details["emulator_plot_path"] = emulator_plot_path

    qare_request_body.update(
        {
            "circuit": circuit_details,
            "ftqc_emulator_emulation": ftqc_emulator_details,
            "layout": {"layout_metadata": layout_metadata},
            "pre": pre_details,
        }
    )
    qare_api_url = f"{API_URLS['qare']}generate_report"
    logging.debug("QARE request body: %s", qare_request_body)

    # Send POST request
    post_qare_response = requests.post(
        qare_api_url, headers=HEADERS, json=qare_request_body
    )
    qare_response = json.loads(post_qare_response.content)
    logging.debug("QARE response: %s", qare_response)

    # Save report to file
    os.makedirs(OUTPUT_DIRECTORY, exist_ok=True)

    # Define the file path where you want to save the JSON content
    file_path = os.path.join(
        OUTPUT_DIRECTORY, f"{circuit_details['circuit_name']}.json"
    )

    # Write the JSON content to a file
    with open(file_path, "w") as file:
        json.dump(qare_response, file, indent=2)

    logging.info("Report saved to %s", file_path)


def run_topqad_e2e(decomposer_circuit_path):
    # Upload circuit that we are testing
    files = {
        "file": (
            decomposer_circuit_path,
            open(decomposer_circuit_path, "rb"),
            "application/octet-stream",
        )
    }

    upload_response = requests.post(API_URLS["upload_url"], files=files)
    scheduler_circuit_path = json.loads(upload_response.content)["fileName"]

    ###### decomposer step ######

    post_decomposer_circuit_path, sk_accumulated_error = run_decomposer(
        scheduler_circuit_path
    )

    ###### optimizer step ######
    transpiled_circuit_path, circuit_details = run_optimizer(
        post_decomposer_circuit_path, sk_accumulated_error
    )

    ###### scheduler step ######
    layout_details, layout_metadata, schedule_details = run_scheduler(
        transpiled_circuit_path
    )

    ###### ftqc_emulator step ######
    (
        ftqc_emulator_request_body,
        emulator_plot_path,
        ftqc_emulator_emulation_details,
        ftqc_emulator_inputs_for_pre,
    ) = run_ftqc_emulator()

    ###### PRE step ######
    pre_details = run_pre(
        circuit_details, layout_details, ftqc_emulator_inputs_for_pre, schedule_details
    )

    ###### QARE step ######
    run_qare(
        circuit_details,
        layout_metadata,
        ftqc_emulator_request_body,
        emulator_plot_path,
        ftqc_emulator_emulation_details,
        pre_details,
    )


if __name__ == "__main__":
    # Check if an argument has been passed
    if len(sys.argv) != 2:
        print("Usage: python benchmarking.py <argument>")
        sys.exit(1)

    # Get the argument passed
    circuit_path = sys.argv[1]

    # Call the main function with the argument
    run_topqad_e2e(circuit_path)
